﻿using System;
using System.Collections.Generic;
using System.IO;
using System.Linq;
using System.Reflection;
using System.Text;
using System.Threading.Tasks;
using NbuLibrary.Core.ModuleEngine;
using NbuLibrary.Core.Services;
using NbuLibrary.Core.Services.tmp;
using Ninject.Modules;
using NbuLibrary.Core.Domain;
using System.Security.Cryptography;
using NbuLibrary.Core.Service.tmp;
using NbuLibrary.Core.DataModel;

namespace NbuLibrary.Core.AccountModule
{
    public class Permissions
    {
        public const string UserActivation = "UserActivation";
        public const string UserManagement = "UserManagement";
        public const string UserGroupManagement = "UserGroupManagement";
    }

    public class AccountNinjectModule : NinjectModule
    {
        public override void Load()
        {
            Bind<IModule>().To<AccountModule>();
            Bind<IEntityOperationLogic>().To<AccountEntityOperationLogic>();
            Bind<IEntityQueryInspector>().To<AccountEntityQueryInspector>();
            Bind<IEntityOperationInspector>().To<AccountEntityOperationInspector>();
        }
    }


    public class AccountModule : IModule
    {
        private IUIProvider _uiProvider;

        public const int Id = 1;
        private IEntityRepository _entityRepository;
        private IDatabaseService _dbService;

        public AccountModule(IEntityRepository entityRepository, IDatabaseService dbService)
        {
            _entityRepository = entityRepository;
            _dbService = dbService;
        }

        public string Name
        {
            get { return "Account"; }
        }

        public IUIProvider UIProvider
        {
            get
            {
                if (_uiProvider == null)
                    _uiProvider = new AccountUIProvider();

                return _uiProvider;
            }
        }

        public void Initialize()
        {
            try
            {
                var q = new EntityQuery2(User.ENTITY);
                q.WhereIs("Email", "admin@nbu.bg");
                using (var dbContext = _dbService.GetDatabaseContext(true))
                {
                    var e = _entityRepository.Search(q).SingleOrDefault();
                    if (e == null)
                    {
                        User admin = new User()
                        {
                            FirstName = "Build in",
                            LastName = "Administrator",
                            Email = "admin@nbu.bg",
                            Password = "adminpass",
                            UserType = UserTypes.Admin,
                            IsActive = true
                        };

                        SHA1 sha1 = SHA1.Create();
                        admin.Password = Convert.ToBase64String(sha1.ComputeHash(Encoding.UTF8.GetBytes(admin.Password)));

                        _entityRepository.Create(admin);
                        dbContext.Complete();
                    }
                }
            }
            catch (Exception ex)
            {
                throw new Exception("Could not create build-in-administrator user.", ex);
            }


            ////If this is a fresh install(no users) create build in admin
            //if (_entityService.Search(new EntityQuery() { EntityName = "User" }).Count() == 0)
            //{
            //    User admin = new User()
            //    {
            //        Email = "admin@nbu.bg", //TODO: Config?
            //        Password = "adminpass", //TODO: Config?
            //        FirstName = "Build In",
            //        LastName = "Administrator",
            //        UserType = UserTypes.Admin,
            //        IsActive = true
            //    };

            //    SHA1 sha1 = SHA1.Create();
            //    admin.Password = Convert.ToBase64String(sha1.ComputeHash(Encoding.UTF8.GetBytes(admin.Password)));


            //}
        }

        int IModule.Id
        {
            get { return AccountModule.Id; }
        }


        public ModuleRequirements Requirements
        {
            get
            {
                ModelBuilder mbUser = new ModelBuilder("User");
                mbUser.AddString("Password", 100);
                mbUser.AddString("Email", 256);
                mbUser.AddBoolean("IsActive", false);
                mbUser.AddString("FirstName", 256);
                mbUser.AddString("MiddleName", 256);
                mbUser.AddString("LastName", 256);
                mbUser.AddComputed("FullName", "[FirstName]+N' '+isnull([MiddleName], N'')+N' '+[LastName]");
                mbUser.AddString("CardNumber", 128);
                mbUser.AddString("FacultyNumber", 128);
                mbUser.AddString("PhoneNumber", 128);
                mbUser.AddEnum<UserTypes>("UserType");
                mbUser.AddUri("URI", "users");

                mbUser.Rules.AddUnique("CardNumber");
                mbUser.Rules.AddUnique("FacultyNumber");
                mbUser.Rules.AddUnique("Email");
                mbUser.Rules.AddRequired("Email");
                mbUser.Rules.AddRequired("Password");
                mbUser.Rules.AddRequired("IsActive");

                ModelBuilder mbGroup = new ModelBuilder("UserGroup");
                mbGroup.AddString("Name", 256);
                mbGroup.AddEnum<UserTypes>("UserType");

                ModelBuilder mbPerm = new ModelBuilder(ModulePermission.ENTITY);
                mbPerm.AddInteger("ModuleID");
                mbPerm.AddString("ModuleName", 256);
                mbPerm.AddString("Available", 1024);

                var rel = mbGroup.AddRelationTo(mbPerm.EntityModel, RelationType.ManyToMany, ModulePermission.DEFAULT_ROLE);
                ModelBuilder mbPermRel = new ModelBuilder(rel);
                mbPermRel.AddString("Granted", 1024, string.Empty);

                mbUser.AddRelationTo(mbGroup.EntityModel, RelationType.ManyToOne, "UserGroup");

                var dm = new DomainModel();
                dm.Entities.Add(mbUser.EntityModel);
                dm.Entities.Add(mbGroup.EntityModel);
                dm.Entities.Add(mbPerm.EntityModel);

                //_definitionService.Ensure(new EntityDefinition[] { user, userGroup, modulepPerm }, new EntityRelation[] { relation, relationGroupPerm });
                return new ModuleRequirements()
                {
                    Domain = dm,
                    //Definitions = new EntityDefinition[] { user, userGroup, modulepPerm },
                    //Relations = new EntityRelation[] { relation, relationGroupPerm },
                    UIDefinitions = GetUIDefs(),
                    Permissions = new string[] { Permissions.UserActivation, Permissions.UserGroupManagement, Permissions.UserManagement }
                };
            }
        }

        private IEnumerable<Service.tmp.UIDefinition> GetUIDefs()
        {
            List<UIDefinition> defs = new List<UIDefinition>();
            defs.Add(new GridDefinition()
            {
                Name = "Account_Admin_UsersGrid",
                Entity = "User",
                Label = "Admin - Users grid",
                Fields = new List<ViewField>()
            });

            defs.Add(new GridDefinition()
            {
                Name = "Account_Admin_UserGroupGrid",
                Entity = "UserGroup",
                Label = "Admin - UserGroups grid",
                Fields = new List<ViewField>()
            });

            defs.Add(new GridDefinition()
            {
                Name = "Account_Admin_PendingRegistrations",
                Entity = "User",
                Label = "Admin - Pending registrations grid",
                Fields = new List<ViewField>()
            });

            defs.Add(new ViewDefinition()
            {
                Name = "Account_Admin_User",
                Entity = "User",
                Label = "Admin - User",
                Fields = new List<ViewField>()
            });

            defs.Add(new FormDefinition()
            {
                Name = "Account_Admin_UserForm",
                Entity = "User",
                Label = "Admin - User Form",
                Fields = new List<EditField>()
            });
            defs.Add(new FormDefinition()
            {
                Name = "Account_Admin_UserGroupForm",
                Entity = "UserGroup",
                Label = "Admin - User Form",
                Fields = new List<EditField>()
            });


            return defs;
        }


        public decimal Version
        {
            get { return 1.6m; }
        }
    }

    public class AccountUIProvider : IUIProvider
    {
        public string GetClientTemplates(Domain.UserTypes type)
        {
            if (type == UserTypes.Admin)
                return GetContent("Templates.AdminUI.html");
            else return string.Empty;
        }

        private string GetContent(string resourceFileName)
        {
            string resourceContent = null;
            using (Stream stream = Assembly.GetExecutingAssembly().GetManifestResourceStream(string.Format("NbuLibrary.Core.AccountModule.{0}", resourceFileName)))
            {
                using (var reader = new StreamReader(stream))
                {
                    resourceContent = reader.ReadToEnd();
                }
            }

            return resourceContent;
        }

        IEnumerable<ClientScript> IUIProvider.GetClientScripts(UserTypes type)
        {
            return new List<ClientScript>() {
                new ClientScript(){
                    Name = "account",
                    Content = GetContent("Scripts.account.js")
                },
                new ClientScript(){
                    Name = "vm.usergroup",
                    Content = GetContent("Scripts.vm.usergroup.js")
                },
                new ClientScript(){
                    Name = "vm.usergroups",
                    Content = GetContent("Scripts.vm.usergroups.js")
                },
                new ClientScript(){
                    Name = "vm.users",
                    Content = GetContent("Scripts.vm.users.js")
                },
                new ClientScript(){
                    Name = "vm.user",
                    Content = GetContent("Scripts.vm.user.js")
                },
                new ClientScript(){
                    Name = "vm.pendingUsers",
                    Content = GetContent("Scripts.vm.pendingUsers.js")
                }
            };
        }
    }

    public class AccountEntityOperationLogic : IEntityOperationLogic
    {
        private const string CTXKEY_PASSWORD_UPDATE = "AccountLogic:PasswordUpdate";
        private const string CTXKEY_USER_CREATION = "AccountLogic:UserCreation";
        private INotificationService _notificationService;

        public AccountEntityOperationLogic(INotificationService notificationService)
        {
            _notificationService = notificationService;
        }

        public void Before(EntityOperation operation, EntityOperationContext context)
        {
            if (operation is EntityUpdate && operation.IsEntity(User.ENTITY))
            {
                var update = operation as EntityUpdate;
                if (update.IsCreate() && !update.ContainsProperty("Password")) //generate random password
                {
                    update.Set("Password", GenerateRandomPassword());
                }

                if (update.ContainsProperty("password"))
                {
                    var newPassword = update.PropertyUpdates["password"] as string;
                    context.Set<string>(CTXKEY_PASSWORD_UPDATE, newPassword);

                    string hash = null;
                    using (SHA1 sha1 = SHA1.Create())
                    {
                        hash = Convert.ToBase64String(sha1.ComputeHash(Encoding.UTF8.GetBytes(newPassword)));
                    }

                    update.Set("password", hash);
                }

                if (update.IsCreate())
                    context.Set<bool>(CTXKEY_USER_CREATION, true);
            }
        }

        public void After(EntityOperation operation, EntityOperationContext context, EntityOperationResult result)
        {
            if (!result.Success)
                return;

            if (context.Get<bool>(CTXKEY_USER_CREATION))
            {
                var update = operation as EntityUpdate;

                var user = new User(update.ToEntity());

                string body = string.Format("Hello, {0} {1},\nAn user was created in the NBU Library's E-Services system with your email.\n", user.FirstName, user.LastName);
                var pwd = context.Get<string>(CTXKEY_PASSWORD_UPDATE);
                if (!string.IsNullOrEmpty(pwd))
                    body = string.Format("{0}Your current password is: \"{1}\".\n", body, pwd);
                if (!update.PropertyUpdates.ContainsKey("isactive") || !user.IsActive)
                    body = string.Format("{0}Your user is not yet active. Once reviewed by staff, you will receive an email.\n", body);

                //TODO: async execution
                _notificationService.SendEmail(user.Email, string.Format("NBU Library user created"), body, null);
            }
        }

        private static string RandomPasswordSymbols = "abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ1234567890";
        private string GenerateRandomPassword()
        {
            Random r = new Random();
            var symbolsCnt = RandomPasswordSymbols.Length;
            var len = r.Next(6, 8);
            var pwd = new char[len];
            for (int i = 0; i < len; i++)
            {
                pwd[i] = RandomPasswordSymbols[r.Next(0, symbolsCnt)];
            }
            return new string(pwd);
        }
    }

    public class AccountEntityOperationInspector : IEntityOperationInspector
    {
        private ISecurityService _securityService;
        private IEntityRepository _repository;
        public AccountEntityOperationInspector(ISecurityService securityService, IEntityRepository repository)
        {
            _securityService = securityService;
            _repository = repository;
        }

        public InspectionResult Inspect(EntityOperation operation)
        {
            if (operation.IsEntity(User.ENTITY))
            {
                if (_securityService.HasModulePermission(_securityService.CurrentUser, AccountModule.Id, Permissions.UserManagement))
                    return InspectionResult.Allow;
                else if (operation.Id.HasValue && _securityService.HasModulePermission(_securityService.CurrentUser, AccountModule.Id, Permissions.UserActivation))
                {
                    EntityQuery2 query = new EntityQuery2("User", operation.Id.Value);
                    query.AddProperty("isActive");
                    var e = _repository.Read(query);
                    if (e.GetData<bool>("IsActive") == false)
                        return InspectionResult.Allow;
                }
                else if (operation is EntityUpdate)
                {
                    var update = operation as EntityUpdate;
                    if (update.Id.HasValue && update.Id.Value == _securityService.CurrentUser.Id)//TODO: allowing the user to edit his data and relations
                        return InspectionResult.Allow;
                }
            }

            return InspectionResult.None;
        }
    }

    public class AccountEntityQueryInspector : IEntityQueryInspector
    {
        private ISecurityService _securityService;
        public AccountEntityQueryInspector(ISecurityService securityService)
        {
            _securityService = securityService;
        }

        public InspectionResult InspectQuery(EntityQuery2 query)
        {
            if (query.IsForEntity(User.ENTITY))
            {
                if (_securityService.HasModulePermission(_securityService.CurrentUser, AccountModule.Id, Permissions.UserManagement))
                    return InspectionResult.Allow;


                bool hasUserActivationPermission = _securityService.HasModulePermission(_securityService.CurrentUser, AccountModule.Id, Permissions.UserActivation);
                var isActiveRule = query.Rules.Find(r => r.IsForProperty("IsActive"));
                if (isActiveRule != null
                    && isActiveRule.Values.Count() == 1
                    && Convert.ToBoolean(isActiveRule.Values.Single()) == false
                    && hasUserActivationPermission)
                    return InspectionResult.Allow;

                if (hasUserActivationPermission && !query.HasProperty("IsActive"))
                    query.AddProperty("IsActive");


            }
            else if (query.IsForEntity(UserGroup.ENTITY))
            {
                if (_securityService.HasModulePermission(_securityService.CurrentUser, AccountModule.Id, Permissions.UserGroupManagement)
                    || _securityService.HasModulePermission(_securityService.CurrentUser, AccountModule.Id, Permissions.UserActivation))
                    return InspectionResult.Allow;
            }

            return InspectionResult.None;
        }

        public InspectionResult InspectResult(EntityQuery2 query, IEnumerable<Entity> entity)
        {
            if (query.IsForEntity(User.ENTITY))
            {
                if (_securityService.HasModulePermission(_securityService.CurrentUser, AccountModule.Id, Permissions.UserManagement))
                    return InspectionResult.Allow;
                else if (_securityService.HasModulePermission(_securityService.CurrentUser, AccountModule.Id, Permissions.UserActivation))
                {
                    foreach (var e in entity)
                    {
                        if (e.GetData<bool>("IsActive") == true)
                            return InspectionResult.None;
                    }
                    return InspectionResult.Allow; //Only inactive users are returned and the user has permission to see inactive users
                }
            }
            else if (query.IsForEntity(UserGroup.ENTITY))
            {
                if (_securityService.HasModulePermission(_securityService.CurrentUser, AccountModule.Id, Permissions.UserGroupManagement))
                    return InspectionResult.Allow;
            }

            return InspectionResult.None;

        }
    }

}
