﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading.Tasks;
using Ninject;
using Ninject.Modules;
using NbuLibrary.Core.Services;
using NbuLibrary.Core.ModuleEngine;
using NbuLibrary.Core.Domain;
using System.Security.Cryptography;
using System.Web.Security;
using NbuLibrary.Core.Services.tmp;
using System.Xml.Serialization;
using NbuLibrary.Core.Service.tmp;
using System.Threading;
using System.IO;
using System.Xml;

namespace NbuLibrary.Core.Infrastructure
{
    public class UIDefinitionServiceImpl : IUIDefinitionService
    {
        private Dictionary<string, UIDefinition> _definitions;
        private string _path;

        private static ReaderWriterLockSlim _rwLock = new ReaderWriterLockSlim();

        public UIDefinitionServiceImpl()
        {
            _path = System.Web.HttpContext.Current.Server.MapPath("~/viewdefinitionstore.xml");

            _rwLock.EnterUpgradeableReadLock();
            try
            {
                if (!System.IO.File.Exists(_path))
                {
                    _rwLock.EnterWriteLock();
                    try
                    {
                        _definitions = new Dictionary<string, UIDefinition>();
                        WriteDefinitionsToFile(FileMode.Create);

                    }
                    finally
                    {
                        _rwLock.ExitWriteLock();
                    }
                }
                else
                {
                    XmlSerializer ser = new XmlSerializer(typeof(List<UIDefinition>));

                    List<UIDefinition> list = null;
                    using (FileStream fs = new FileStream(_path, FileMode.Open))
                    {
                        list = ser.Deserialize(fs) as List<UIDefinition>;
                    }
                    _definitions = new Dictionary<string, UIDefinition>();
                    list.ForEach(uidef => _definitions.Add(uidef.Name, uidef));
                }
            }
            finally
            {
                _rwLock.ExitUpgradeableReadLock();
            }
        }

        public Service.tmp.UIDefinition GetByName(string name)
        {
            _rwLock.EnterReadLock();
            UIDefinition def = null;
            try
            {
                _definitions.TryGetValue(name, out def);
            }
            finally
            {
                _rwLock.ExitReadLock();
            }

            return def;
        }

        public T GetByName<T>(string name) where T : Service.tmp.UIDefinition
        {
            _rwLock.EnterReadLock();
            UIDefinition def = null;
            try
            {
                _definitions.TryGetValue(name, out def);
            }
            finally
            {
                _rwLock.ExitReadLock();
            }

            return def as T;
        }

        public void Add(Service.tmp.UIDefinition definition)
        {
            PrepareDefinition(definition);
            _rwLock.EnterWriteLock();
            try
            {
                _definitions.Add(definition.Name, definition);
                WriteDefinitionsToFile(FileMode.Truncate);

            }
            finally
            {
                _rwLock.ExitWriteLock();
            }
        }

        public void Remove(Service.tmp.UIDefinition definition)
        {
            _rwLock.EnterWriteLock();
            try
            {
                _definitions.Remove(definition.Name);
                WriteDefinitionsToFile(FileMode.Truncate);

            }
            finally
            {
                _rwLock.ExitWriteLock();
            }
        }

        public void Update(Service.tmp.UIDefinition definition)
        {
            PrepareDefinition(definition);
            _rwLock.EnterWriteLock();
            try
            {
                if (!_definitions.ContainsKey(definition.Name))
                    throw new ArgumentException(string.Format("The definition {0} does not exist!", definition.Name));

                _definitions[definition.Name] = definition;
                WriteDefinitionsToFile(FileMode.Truncate);
            }
            finally
            {
                _rwLock.ExitWriteLock();
            }
        }

        public IEnumerable<Service.tmp.UIDefinition> GetAll()
        {
            _rwLock.EnterReadLock();
            IEnumerable<UIDefinition> all = null;
            try
            {
                all = _definitions.Values.OrderBy(f => f.Name).ToArray();
            }
            finally
            {
                _rwLock.ExitReadLock();
            }

            return all;
        }

        private void WriteDefinitionsToFile(FileMode mode)
        {
            foreach (var def in _definitions.Values)
                PrepareDefinition(def);
            var nf = string.Format("{0}_new", _path);
            System.IO.File.Copy(_path, nf);
            XmlSerializer ser = new XmlSerializer(typeof(List<UIDefinition>));
            try
            {
                using (FileStream fs = new FileStream(nf, mode))
                {
                    ser.Serialize(fs, _definitions.Values.ToList());
                    fs.Flush();
                }
                System.IO.File.Copy(nf, _path, true);
            }
            finally {
                System.IO.File.Delete(nf);
            }
        }

        private void PrepareDefinition(UIDefinition definition)
        {
            if (definition is GridDefinition)
            {
                var gridDef = definition as GridDefinition;
                if (gridDef.Fields == null)
                    gridDef.Fields = new List<ViewField>();
                gridDef.Fields = gridDef.Fields.OrderBy(f => f.Order).ToList();
                //TODO: Filters
            }
            else if (definition is ViewDefinition)
            {
                var viewDef = definition as ViewDefinition;
                if (viewDef.Fields == null)
                    viewDef.Fields = new List<ViewField>();
                viewDef.Fields = viewDef.Fields.OrderBy(f => f.Order).ToList();
            }
            else if (definition is FormDefinition)
            {
                var formDef = definition as FormDefinition;
                if (formDef.Fields == null)
                    formDef.Fields = new List<EditField>();
                formDef.Fields = formDef.Fields.OrderBy(f => f.Order).ToList();
            }
        }
    }


    public class AdminBusinessLogic : IBusinessLogic
    {
        private ISecurityService _securityService;
        public AdminBusinessLogic(ISecurityService securityService)
        {
            _securityService = securityService;
        }

        public OperationResolution Resolve(EntityOperation operation)
        {
            if (_securityService.CurrentUser.UserType == UserTypes.Admin)
                return OperationResolution.Allowed;
            else
            {
                return OperationResolution.Allowed;
            }
            return OperationResolution.None;
        }


        public void Prepare(EntityOperation operation, ref BLContext context)
        {
        }

        public void Complete(EntityOperation operation, ref BLContext context, ref BLResponse response)
        {
        }
    }

   

}
